# All sympy stuff is hidden here for faster code parsing
from typing import Callable, Union
from sympy import Symbol, parse_expr, lambdify, diff, Mul, Add, Pow, sqrt, latex, init_printing

init_printing()

def list_swap(l: list, x, y):
    if x not in l or y not in l:
        ValueError("elements not in given list!")
    el = [l.index(x), l.index(y)]
    l[el[0]], l[el[1]] = l[el[1]], l[el[0]]
text = ["ab", "a", "b", "bc", "c"]

def sort_string_list(l: list[str]) -> list[str]: 
    out = l
    for x in out:
        for y in out:
            if x in y and out.index(x) < out.index(y):
                list_swap(text, x, y)
    return out

def get_function(function: str, vars: list[str], consts: list[str] = [], error_symbol: str = r"\Delta") -> Union[Callable, str, Callable, str]:
    func = function
    all_text = sort_string_list(vars + consts)
    hashes = [str(hash(var)) for var in all_text]
    symbols = [Symbol(var) for var in all_text]
    symbols_vars = [Symbol(var) for var in vars] # Only variables
    for var, h in zip(all_text, hashes):
         func = func.replace(var, h)
    for var, h in zip(all_text, hashes):
         func = func.replace(h, f"Symbol('{var}')")
    func = parse_expr(func)
    out_func = lambdify(symbols,func, "numpy")

    errors = [Symbol(err) for err in [error_symbol + " " + var for var in vars]]
    lambdify_errs = [Symbol(err) for err in [a + "_err" for a in vars]] # Necessary because variables with "\Delta" are not interpreted properly
    err_func_latex = 0
    err_func_lambdify = 0
    for i in range(len(errors)): 
        err_func_latex = Add(err_func_latex, Pow(Mul(errors[i], diff(func, symbols_vars[i]), evaluate= False), 2, evaluate=False))
        err_func_lambdify = Add(err_func_lambdify, Pow(Mul(lambdify_errs[i], diff(func, symbols_vars[i]), evaluate= False), 2, evaluate=False))
    err_func_latex = sqrt(err_func_latex)
    out_err_func = lambdify(symbols + lambdify_errs, err_func_lambdify, "numpy")
    return out_func, latex(func), out_err_func, latex(err_func_latex)



from . import __define_unit
from ..units import Quantity

quetta = Quantity(1e30, units=1, custom_string="Q", expect_self=True)
ronna  = Quantity(1e27, units=1, custom_string="R", expect_self=True)
yotta  = Quantity(1e24, units=1, custom_string="Y", expect_self=True)
zetta  = Quantity(1e21, units=1, custom_string="Z", expect_self=True)
exa    = Quantity(1e18, units=1, custom_string="E", expect_self=True)
peta   = Quantity(1e15, units=1, custom_string="P", expect_self=True)
tera   = Quantity(1e12, units=1, custom_string="T", expect_self=True)
giga   = Quantity(1e9, units=1, custom_string="G", expect_self=True)
mega   = Quantity(1e6, units=1, custom_string="M", expect_self=True)
kilo   = Quantity(1e3, units=1, custom_string="k", expect_self=True)
hecto  = Quantity(1e2, units=1, custom_string="h", expect_self=True)
deca   = Quantity(1e1, units=1, custom_string="da", expect_self=True)
deci   = Quantity(1e-1, units=1, custom_string="d", expect_self=True)
centi  = Quantity(1e-2, units=1, custom_string="c", expect_self=True)
milli   = Quantity(1e-3, units=1, custom_string="m", expect_self=True)
micro  = Quantity(1e-6, units=1, custom_string="µ", expect_self=True)
nano   = Quantity(1e-9, units=1, custom_string="n", expect_self=True)
pico   = Quantity(1e-12, units=1, custom_string="p", expect_self=True)
femto  = Quantity(1e-15, units=1, custom_string="f", expect_self=True)
atto   = Quantity(1e-18, units=1, custom_string="a", expect_self=True)
zepto  = Quantity(1e-21, units=1, custom_string="z", expect_self=True)
yocto  = Quantity(1e-24, units=1, custom_string="y", expect_self=True)
ronto  = Quantity(1e-27, units=1, custom_string="r", expect_self=True)
quecto = Quantity(1e-30, units=1, custom_string="q", expect_self=True)



from .sympy_enclosure import get_function
from .units import Quantity, remove_units
from IPython.display import display, Latex, Markdown
from pandas import Series, DataFrame
from pandas.api.types import is_list_like
from typing import Union

def identify_error_symbol(indices: list[str] | str, possible_symbols: list[str]) -> str:
    if isinstance(indices, str):
        indices = [indices]
    error_symbol = ""
    for symbol in possible_symbols:
        if any(symbol in e for e in indices):
            error_symbol = symbol
            break
    return error_symbol

class Helper:
    _possible_error_symbols = ["_err", "err", "Δ"]
    def __init__(self, function: str,vars_in: list[str] = [], const_in: list[str] = [], error_mark: str = r"\Delta") -> None:
        for var in vars_in + const_in:
            if var not in function:
                raise ValueError(f"{var} not in function")
        self._vars = vars_in
        self._consts = const_in
        self._values = {} # TODO: mabye switch this to DataFrame?
        self._function, self._function_latex, self._error_function, self._error_function_latex = get_function(function, vars_in, const_in, error_mark)

    def get_inputs(self) -> list[str]:
        return list(self._function.__code__.co_varnames)

    def get_error_inputs(self, include_regular_inputs: bool = True, clean: bool = False) -> list[str]:
        errs_only = [e for e in self._error_function.__code__.co_varnames if e not in self.get_inputs()] 
        if clean:
            errs_only = [a.split("_err")[0] for a in self.get_error_inputs(include_regular_inputs=False)]
        if include_regular_inputs:
            return self.get_inputs() + errs_only
        else:
            return errs_only

    def _find_missing_indices(self, d: Series | DataFrame) -> list[str]:
        indices = list(d.index) if isinstance(d, Series) else list(d.columns)
        for key, value in self._values.items():
            if key not in indices:
                d[key] = value
        indices = list(d.index) if isinstance(d, Series) else list(d.columns)
        missing = [a for a in self.get_inputs() if a not in indices]
        return missing

    def _find_missing_error_indices(self, d: Series | DataFrame) -> tuple[list[str], list[str]]:
        indices = list(d.index) if isinstance(d, Series) else list(d.columns)
        # Rename possible error symbols to "_err" for consistency
        error_symbol = identify_error_symbol(indices, Helper._possible_error_symbols)
        if error_symbol != "":
            error_indices = [e for e in indices if error_symbol in e]
            valid_error_indices = [e.replace(error_symbol, "", 1) + "_err" for e in error_indices]
            rename_map = {invalid: valid for invalid, valid in zip(error_indices, valid_error_indices)}
            d.rename(columns=rename_map, inplace=True)
            indices = list(d.index) if isinstance(d, Series) else list(d.columns)
        
        for key, value in self._values.items():
            if key not in indices:
                d[key] = value
        indices = list(d.index) if isinstance(d, Series) else list(d.columns)
        return [a for a in self.get_error_inputs() if a not in indices]

    def _validate_key(self, keys) -> list[str]:
        if isinstance(keys, str):
            keys = [keys]
        invalid_types = []
        for key in keys:
            if not isinstance(key, str):
                invalid_types.append(type(key))
        if invalid_types:
            raise ValueError(f"Indexing is only allowed through strings, not {invalid_types}")
        
        if all(key in self.get_inputs() for key in keys):
            return keys
        
        symbol = identify_error_symbol(keys, Helper._possible_error_symbols)
        if symbol == "":
            raise ValueError(f"Key is not valid. Reminder: possible error symbols are {Helper._possible_error_symbols}")
        non_errors = [key for key in keys if symbol not in key]
        errors = [key.replace(symbol, "", 1) for key in keys if key not in non_errors]
        valid_errors = [key + "_err" for key in errors]
        final = non_errors + valid_errors
        for key in final:
            if key not in self.get_error_inputs():
                raise ValueError(f"Key {key} is not one of {self.get_error_inputs()}")
        return final
        
    def __getitem__(self, keys):
        keys = self._validate_key(keys)
        not_set = [key for key in keys if key not in self._values.keys()]
        if not_set:
            raise ValueError(f"Key {not_set} has not been set")
        
        return [self._values[key] for key in keys] if len(keys) > 1 else self._values[keys[0]]
    def __setitem__(self, keys, values):
        keys = self._validate_key(keys)
        if not is_list_like(values):
            values = [values]
        self._values |= {key: value for key, value in zip(keys, values)}

    def __call__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], Series) or isinstance(args[0], DataFrame):
            data = args[0]
            missing = self._find_missing_indices(data)
            if missing:
                raise ValueError(f"Required inputs {missing} not in Series/DataFrame")
            inputs = data[self.get_inputs()].values.T
        elif isinstance(args, tuple):
            if len(args) != len(self.get_inputs()):
                raise ValueError(f"Number of inputs ({len(args)}) does not match required number of inputs ({len(self.get_inputs())}, {self.get_inputs()})")
            inputs = args
        return self._function(*inputs)

    def error(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], Series) or isinstance(args[0], DataFrame):
            data = args[0]
            missing = self._find_missing_error_indices(data)
            if missing:
                raise ValueError(f"Required inputs {missing} not in Series/DataFrame")
            inputs = data[self.get_error_inputs()].map(remove_units).values.T # Remove units for now since unit**1/2 is not supported
            final = self._error_function(*inputs) # In SI units
            units = self._function(*data[self.get_inputs()].iloc[0].values)
            if isinstance(units, Quantity):
                units, expected_units = units.units, units.expected_units
                print(final)
                final = [Quantity(value=a, units=units, expected_units=expected_units) for a in final] if is_list_like(final) else Quantity(value=final, units=units, expected_units=expected_units)
            return final
        elif isinstance(args, tuple):
            if len(args) != len(self.get_error_inputs()):
                raise ValueError(f"Number of inputs ({len(args)}) does not match required number of inputs ({len(self.get_error_inputs())}, {self.get_error_inputs()})")
            inputs = args
            return self._error_function(*inputs)


    def __repr__(self):
        print("Variables:")
        display(Markdown(", ".join([f"${var}$" for var in self._vars])))
        print("Constants:")
        display(Markdown(", ".join([f"${var}$" for var in self._consts])))
        print("Function:")
        display(Latex(f"${self._function_latex}$"))
        print("Error Function:")
        display(Latex(f"${self._error_function_latex}$"))
        return ""


from numbers import Number
from enum import IntEnum
from typing import Callable
from fractions import Fraction
from math import sqrt, prod, ceil, floor, trunc, log

def prime_factorization_int(n: int) -> list[tuple[int, int]]:
    if n == 1:
        return []
    elif n == 0:
        return []

    fac = []
    powers = []
    count = 0
    while n % 2 == 0:
        count += 1
        n = n // 2
    fac.append(2) if count > 0 else 0
    powers.append(count) if count > 0 else 0
    # n must be odd at this point
    # so a skip of 2 ( i = i + 2) can be used
    for i in range(3,int(sqrt(n))+1,2):
        # while i divides n , add i to the list
        count = 0
        while n % i== 0:
            count += 1
            n = n // i
        fac.append(i) if count > 0 else 0
        powers.append(count) if count > 0 else 0
    # Condition if n is a prime
    # number greater than 2
    if n > 2:
        fac.append(n)
        powers.append(1) 
    return [(factor, power) for factor, power in zip(fac, powers)]


def prime_factorization(n: Fraction) -> list[tuple[int, int]]:
    positive_factors = []
    negative_factors = []

    positive_factors = prime_factorization_int(n.numerator)
    negative_factors = [(factor, -power) for factor, power in prime_factorization_int(n.denominator)]
    final = positive_factors + negative_factors
    final.sort()
    return final

def custom_factors(n: Fraction, custom_factors: list[Fraction]): # TODO: Check this because it's probably wildly unoptimized
    custom_factors = [e for e in custom_factors if e != 1]
    num_factors = prime_factorization(n)
    factors_check = [prime_factorization(custom) for custom in custom_factors]
    final = []
    for fac, custom in zip(factors_check, custom_factors):
        #if all(prime_fac[0] in [e[0] for e in num_factors] for prime_fac in fac):
        not_matching = [a for a in fac if a[0] not in [b[0] for b in num_factors] ]
        if not_matching:
            final.append((custom, 1))
            common_num = [a for a in num_factors if a[0] in [b[0] for b in fac]]
            different_num = [a for a in num_factors if a not in common_num]
            common_fac = [a for a in fac if a[0] in [b[0] for b in num_factors]]
            different_fac = [a for a in fac if a not in common_fac]
            common = [(a[0], a[1] - b[1]) for a, b in zip(common_num, common_fac)]
            num_factors = common + different_num + [(a[0], -a[1]) for a in different_fac]
            continue
            
        matching = [a for a in num_factors if a[0] in [b[0] for b in fac]]
        not_matching = [a for a in num_factors if a not in matching]
        power = min([prime_n[1]//prime_fac[1] for prime_fac, prime_n in zip(fac, matching)], key=abs)
        if power == 0:
            num_factors = [(prime_n[0], prime_n[1] - prime_fac[1]) for prime_fac, prime_n in zip(fac, matching) if prime_n[1] - prime_fac[1] != 0] + not_matching
            final.append((custom, 1))
        else:
            num_factors = [(prime_n[0], prime_n[1] - prime_fac[1]*power) for prime_fac, prime_n in zip(fac, matching) if prime_n[1] - prime_fac[1]*power != 0] + not_matching
            final.append((custom, power))
    return [e for e in final + num_factors if e[1] != 0]

class Quantity:
    _SI_map: dict[int, str] = {2: "m", 3:"s", 5:"kg", 7:"K", 11:"A", 13:"mol", 17:"cd"}
    def __init__(self, value: Number = 1, units: Fraction = 1, expected_units: list = [], custom_string: str = "", expect_self: bool = False) -> None:
        if not isinstance(units, Fraction):
            units = Fraction(units)
        self.units: Fraction = units
        self.value = value
        self.expected_units: list[Quantity] = expected_units
        self.custom_string: str = custom_string
        if expect_self:
            self.expected_units = [self]

    def _units_to_strings(self) -> tuple[Number, str]:
        units_vals = self.units
        value = self.value
        custom_map = {unit.units: unit.custom_string for unit in self.expected_units} 
        custom_map |= {key: val for key, val in Quantity._SI_map.items() if key not in custom_map}
        units = []
        factors = custom_factors(units_vals, [e.units for e in self.expected_units])
        for unit in self.expected_units:
            try:
                power = [e[1] for e in factors if e[0] == unit.units][0]
            except:
                continue
            value /= unit.value**power

        units = [f"{custom_map.get(e[0])}^{e[1]}" if e[1] != 1 else f"{custom_map.get(e[0])}" for e in factors]
        return value, " * ".join(units)

    def _set_custom_string(self, text: str) -> None:
        self.custom_string = text

    def _get_expected_units(self, other, division: bool = False) -> list | None:
        if not self.expected_units or not other.expected_units:
            return self.expected_units + other.expected_units

        if self.expected_units[0].units == 1 or other.expected_units[0].units == 1:
            unit1, unit2 = self.expected_units[0], other.expected_units[0]
            newUnit = Quantity(value=unit1.value*unit2.value, units=unit1.units*unit2.units)
            newUnit.custom_string = unit1.custom_string + unit2.custom_string
            return [newUnit]
        final_self = [a for a in self.expected_units if a.units not in [e.units for e in other.expected_units]]
        if division:
            return final_self
        return final_self + other.expected_units
    
    def __str__(self, significant_digits: int = 3): 
        val, units = self._units_to_strings() 
        modifier = "%." + str(significant_digits) + "g"
        return f"{modifier % val} {units}"

    def __repr__(self):
        return str(self)

    # Multiplication
    
    def __mul__(self, other):
        if isinstance(other, Number):
            return Quantity(value=self.value*other, units=self.units, expected_units=self.expected_units, custom_string=self.custom_string)
        if isinstance(other, Quantity):
            return Quantity(value=self.value*other.value, units=self.units*other.units, expected_units=self._get_expected_units(other))

    def __rmul__(self, other):
        if isinstance(other, Number):
            return Quantity(value=self.value*other, units=self.units, expected_units=self.expected_units, custom_string=self.custom_string)

    def __truediv__(self, other):
        if isinstance(other, Number):
            return Quantity(value=self.value/other, units=self.units, custom_string=self.custom_string, expected_units=self.expected_units)
        if isinstance(other, Quantity):
            return Quantity(value=self.value/other.value, units=self.units/other.units, expected_units=self._get_expected_units(other, division=True))

    def __rtruediv__(self, other):
        if isinstance(other, Number):
            return Quantity(value=other/self.value, units=1/self.units, custom_string=self.custom_string, expected_units=self.expected_units)

    def __floordiv__(self, other):
        if isinstance(other, Number):
            return Quantity(value=self.value//other, units=self.units, custom_string=self.custom_string, expected_units=self.expected_units)
        if isinstance(other, Quantity):
            return Quantity(value=self.value//other.value, units=self.units/other.units, expected_units=self._get_expected_units(other, division=True))
    def __rfloordiv__(self, other):
        if isinstance(other, Number):
            return Quantity(value=other//self.value, units=1/self.units, custom_string=self.custom_string, expected_units=self.expected_units)
    def __mod__(self, other):
        if isinstance(other, Number):
            return Quantity(value=self.value%other, units=self.units, custom_string=self.custom_string, expected_units=self.expected_units)
        if isinstance(other, Quantity):
            return Quantity(value=self.value%other.value, units=self.units/other.units, expected_units=self._get_expected_units(other))
    def __rmod__(self, other):
        if isinstance(other, Number):
            return Quantity(value=other%self.value, units=1/self.units, custom_string=self.custom_string, expected_units=self.expected_units)

    def __pow__(self, other):
        if isinstance(other, int):
            return Quantity(value=self.value**other, units=self.units**other, expected_units=self.expected_units)

    # Comparison
    def __lt__(self, other):
        if isinstance(other, Quantity):
            if other.units == self.units:
                return self.value < other.value
            else:
                raise ValueError("Units do not match!")
        else:
            raise ValueError("Comparison is only valid between Quantities")
    def __gt__(self, other):
        return not self < other
    def __eq__(self, other):
        if isinstance(other, Quantity):
            if other.units == self.units:
                return self.value == other.value
            else:
                raise ValueError("Units do not match!")
        else:
            raise ValueError("Comparison is only valid between Quantities")
    def __le__(self, other):
        return self < other or self == other
    def __ge__(self, other):
        return self > other or self == other
    def __ne__(self, other):
        return not self == other

    
    # Addition
    def __neg__(self):
        return Quantity(value=-self.value, units=self.units, custom_string=self.custom_string, expected_units=self.expected_units)
    def __add__(self, other):
        if isinstance(other, Quantity):
            if self.units != other.units:
                raise ValueError("Units do not match!")
            return Quantity(value=self.value+other.value, units=self.units, expected_units=self._get_expected_units(other))
        else:
            raise ValueError("Quantities can only be added with other quantities")
    def __radd__(self, other):
        if not isinstance(other, Quantity):
            raise ValueError("Quantities can only be added with other quantities")
    def __sub__(self, other):
        return self.__add__(-other)

    def __float__(self):
        return float(self.value)
    
    # Rounding

    def __round_general__(self, method: Callable, decimals: int | None = None):
        value = self.value
        value /= prod([e.value for e in self.expected_units])
        rounded = method(value, decimals) if decimals is not None else method(value)
        value = rounded * prod([e.value for e in self.expected_units])
        return Quantity(value=value, units=self.units, custom_string=self.custom_string, expected_units=self.expected_units)
    def __round__(self, decimals = 0): return self.__round_general__(round, decimals)
    def __ceil__(self): return self.__round_general__(ceil)
    def __floor__(self): return self.__round_general__(floor)
    def __trunc__(self): return self.__round_general__(trunc)

    # For numpy support:

    def sqrt(self): return sqrt(self.value)
    def rint(self): return self.__round__()
    def log(self): return log(self.value)
    
def work_on_units(func):
    def wrapper(*args, **kwargs):
        if args:
            args = list(args)
            arg = args.pop(0)
            if isinstance(arg, Quantity):
                final = func(arg, *args, **kwargs) if args else func(arg, **kwargs)
            elif isinstance(arg, list):
                final = []
                for a in arg:
                    final.append(wrapper(a, *args, **kwargs) if args else wrapper(a, **kwargs))
            else:
                final = arg
            return final
        else:
            arg = kwargs.pop(list(kwargs.keys())[0])
            if isinstance(arg, Quantity):
                final = func(arg, **kwargs) if args else func(arg)
            elif isinstance(arg, list):
                final = []
                for a in arg:
                    final.append(wrapper(a, **kwargs) if args else wrapper(a, **kwargs))
            else:
                final = arg
            return final
    return wrapper
@work_on_units
def remove_units(x: Quantity):
    val, units = x._units_to_strings()
    return val

@work_on_units
def to_SI(x: Quantity):
    return Quantity(x.value, x.units, [], custom_string = x.custom_string)

@work_on_units
def to_units(x, units):
    if not isinstance(units, list):
        units = [units]
    final = []
    for unit in units:
        final += unit.expected_units
    expected_units = final
    return Quantity(x.value, x.units, expected_units, x.custom_string)



from .stats import (random_error_of_mean, student_t_n, coefficient_errors, bootstrap_points)
from .pandas_utils import (df_switch_columns, df_switch_rows, df_create, copy_to_clipboard, df_to_latex, multiindex_df)
from .symbolic import (Helper) 
from .units import (remove_units, to_SI, to_units)
from .graphing import (set_style)



from seaborn import set_style as ss

def set_style():
    ss(style="darkgrid", rc={"axes.grid": True, "xtick.bottom": True, "ytick.left": True, "axes.edgecolor": ".15", "axes.spines.right": False, "axes.spines.top": False})




from . import __define_unit
from .prefixes import *
from ..units import Quantity, DefaultUnits

centimetre = Quantity(1e-2, DefaultUnits.meter, custom_string="cm", expect_self=True)
gram = Quantity(1e-3, DefaultUnits.kilogram, custom_string="g", expect_self=True)
second = Quantity(1, DefaultUnits.second, custom_string="s", expect_self=True)
gal = centimetre/second**2
__define_unit(gal, "Gal")
dyne = gram*centimetre/second**2
__define_unit(dyne, "dyn")
erg = gram * centimetre**2/second**2
__define_unit(erg, "erg")
barye = gram/(centimetre*second**2)
__define_unit(barye, "Ba")
poise = gram/(centimetre*second)
__define_unit(poise, "P")
stokes = centimetre**2/second
__define_unit(stokes, "St")
kayser = centimetre**-1
__define_unit(kayser, "K")




from . import __define_unit
from .prefixes import *
from ..units import Quantity

meter = Quantity(1, 2, custom_string="m", expect_self=True)
second = Quantity(1, 3, custom_string="s", expect_self=True)
kilogram = Quantity(1, 5, custom_string="kg", expect_self=True)
gram = Quantity(1e-3, 5, custom_string="g", expect_self=True)
kelvin = Quantity(1, 7, custom_string="K", expect_self=True)
ampere = Quantity(1, 11, custom_string="A", expect_self=True)
mol = Quantity(1, 13, custom_string="mol", expect_self= True)
candela = Quantity(1, 17,custom_string="cd", expect_self=True)

hertz = second**-1
__define_unit(hertz, "Hz")
newton = kilogram*meter*second**-2
__define_unit(newton, "N")
pascal = newton/meter**2
__define_unit(pascal, "Pa")
atmosphere = 101325 * pascal
__define_unit(atmosphere, "atm")
bar = 100000 * pascal
__define_unit(bar, "bar")
joule = newton*meter
__define_unit(joule, "J")
watt = joule/second
__define_unit(watt, "W")
coulomb = ampere*second
__define_unit(coulomb, "C")
volt = joule/coulomb
__define_unit(volt, "V")
electronvolt = 1.6e-19*joule
__define_unit(electronvolt, "eV")
farad = coulomb/volt
__define_unit(farad, "F")
ohm = volt/ampere
__define_unit(ohm, "Ω")
siemens = ohm**-1
__define_unit(siemens, "S")
weber = volt*second
__define_unit(weber, "Wb")
tesla = weber/meter**2
__define_unit(tesla, "T")
henry = weber/ampere
__define_unit(henry, "H")
# TODO: ºC

lumen = 1*candela
__define_unit(lumen, "lm")
lux = lumen/meter**2
__define_unit(lux, "lx")
becquerel = second**-1
__define_unit(becquerel, "Bq")
gray = joule/kilogram
__define_unit(gray, "Gy")
sievert = joule/kilogram
__define_unit(sievert, "Sv")
katal = mol*second**-1
__define_unit(katal, "kat")

liter = (deci*meter)**3
__define_unit(liter, "L")
